/***************************************************************************
 *   Copyright (C) 2004 by Antonio Fasolato                                *
 *   neo@localhost                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "convertadvance.h"

#include <qfileinfo.h>

#include <kprogress.h>
#include <ksqueezedtextlabel.h>
#include <klocale.h>
#include <kmessagebox.h>

ConvertAdvance::ConvertAdvance(QWidget* parent, const char* name, WFlags fl)
	: ConvertAdvanceBase(parent,name,fl), initOK(true), finished(false), currentTotalSize(0), maxSize(0), chrono(), outputFile(),  converting(false) {
	if(!gst_init_check(NULL,NULL)) {
		KMessageBox::error(0, i18n("<h3>Gstreamer could not be initialized</h3><p>Please make sure that you have installed all necessary GStreamer plugins (e.g. OGG and MP3), and run <i>'gst-register'</i> afterwards.</p><p>For further assistance consult the GStreamer manual, and join #gstreamer on irc.freenode.net.</p>"), i18n("Could not initialize GStreamer"));
		initOK=false;
	}
	
	GstElement* dummy = gst_element_factory_make ( "fakesink", "fakesink" );
	if(!dummy) {
		KMessageBox::error(0, i18n("<h3>Gstreamer misses a register</h3><p>Please make sure that you have installed all necessary GStreamer plugins (e.g. OGG and MP3), and run <i>'gst-register'</i> afterwards.</p><p>For further assistance consult the GStreamer manual, and join #gstreamer on irc.freenode.net.</p>"), i18n("GStreamer misses a register"));
		initOK=false;
	}

	errors=new Error(this);
	
	timer=new QTimer(this);
	chrono.start();
	connect(timer,SIGNAL(timeout()),this,SLOT(tick()));
}

ConvertAdvance::~ConvertAdvance() {
}

void ConvertAdvance::setOptions( Options * o ) {
	options=o;
	finished=false;
	maxSize=0;
	currentTotalSize=0;
}

void ConvertAdvance::startWorking( ) {
	
	QStringList l=options->getFileList();
	QStringList tmp;    //new file list without not existing files (or 0 bytes files) and unreadable files.
	
	//get total work length
	for(QStringList::Iterator i=l.begin(); i!=l.end(); ++i) {
		QFileInfo info(*i);
		if(info.size()!=0 && info.isReadable()) {
			maxSize+=info.size();
			tmp+=*i;
		}
		else {
			errors->setError(*i,ERROR_FILE_NOT_EXISTS);
		}
	}
	
	totalProgress->setTotalSteps(maxSize);

	options->setFileList(tmp);
	
	if(options->getFormat()==wav) {
		errors->setError(i18n("All output files"),ERROR_NO_TAGS);  //wav format has no tags
		options->setMisc(Options::OPTION_TRANSFER_TAGS,false);  //no use trying to work with tags
	}

	currentFile=0;
	converting=true;
	processFile();
}

void ConvertAdvance::processFile( ) {
	if(currentFile!=options->getFileList().size()) {
		//find sourcefile format (based on file extension)
		if((options->getFileList())[currentFile].endsWith(".mp3") || (options->getFileList())[currentFile].endsWith(".MP3"))
			options->setInputFormat(mp3);
		else if((options->getFileList())[currentFile].endsWith(".ogg") || (options->getFileList())[currentFile].endsWith(".OGG"))
			options->setInputFormat(ogg);
		else if((options->getFileList())[currentFile].endsWith(".wav") || (options->getFileList())[currentFile].endsWith(".WAV")) {
			options->setInputFormat(wav);
			if(options->getMisc(Options::OPTION_TRANSFER_TAGS))
				errors->setError(options->getFileList()[currentFile],ERROR_NO_TAGS);  //wav format has no tags
		}
		else {
			errors->setError(options->getFileList()[currentFile],ERROR_FORMAT_UNKNOWN);
			abortedOperation(QFileInfo((options->getFileList())[currentFile]).size());
			return;
		}
		
		if(options->getInputFormat()!=wav) {
			TagLib::FileRef in((options->getFileList())[currentFile].ascii());
			if(!in.isNull() && !(in.tag()->isEmpty())) {
				authorLabel->setText((in.tag()->artist()).toCString());
				titleLabel->setText((in.tag()->title()).toCString());
				albumLabel->setText((in.tag()->album()).toCString());
				yearLabel->setText(QString::number(in.tag()->year()));
			}
		}
		
		outputFile=createOutputFilename();
		if(outputFile=="") {   //file exists, we cannot overwrite it
			abortedOperation(QFileInfo((options->getFileList())[currentFile]).size());
			return;
		}
		
		inputLabel->setText((options->getFileList())[currentFile]);
		outputLabel->setText(outputFile);
		
		//Creating the pipeline and its elements
		pipeline=gst_thread_new("pipeline");
		g_signal_connect(G_OBJECT(pipeline), "eos", G_CALLBACK(ConvertAdvance::eosCb),this);

		fileSrc=gst_element_factory_make("filesrc","disk_source");
		if(fileSrc==NULL) {
			errors->setError(options->getFileList()[currentFile],ERROR_GST_INTERNAL);
			abortedOperation(QFileInfo((options->getFileList())[currentFile]).size());
			return;
		}
		gst_element_set(fileSrc, "location", (options->getFileList())[currentFile].ascii(), NULL);
		
		spider=gst_element_factory_make("spider","spdr");
		if(spider==NULL) {
			errors->setError(options->getFileList()[currentFile],ERROR_GST_INTERNAL);
			abortedOperation(QFileInfo((options->getFileList())[currentFile]).size());
			return;
		}
		
		//output format specific
		switch(options->getFormat()) {
			case mp3: {
				encoder=gst_element_factory_make("lame","encode");
				if(encoder==NULL) {
					errors->setError(options->getFileList()[currentFile],ERROR_GST_INTERNAL);
					abortedOperation(QFileInfo((options->getFileList())[currentFile]).size());
					return;
				}
				fileSink=gst_element_factory_make("filesink","disk_sink");
				if(fileSink==NULL) {
					errors->setError(options->getFileList()[currentFile],ERROR_GST_INTERNAL);
					abortedOperation(QFileInfo((options->getFileList())[currentFile]).size());
					return;
				}
				gst_element_set(fileSink, "location", outputFile.ascii(), NULL);
				break;
			}
			case ogg: {
				encoder=gst_element_factory_make("vorbisenc","encode");
				if(encoder==NULL) {
					errors->setError(options->getFileList()[currentFile],ERROR_GST_INTERNAL);
					abortedOperation(QFileInfo((options->getFileList())[currentFile]).size());
					return;
				}
				fileSink=gst_element_factory_make("filesink","disk_sink");
				if(fileSink==NULL) {
					errors->setError(options->getFileList()[currentFile],ERROR_GST_INTERNAL);
					abortedOperation(QFileInfo((options->getFileList())[currentFile]).size());
					return;
				}
				gst_element_set(fileSink, "location", outputFile.ascii(), NULL);
				break;
			}
			case wav: {
				encoder=gst_element_factory_make("wavenc","encode");
				if(encoder==NULL) {
					errors->setError(options->getFileList()[currentFile],ERROR_GST_INTERNAL);
					abortedOperation(QFileInfo((options->getFileList())[currentFile]).size());
					return;
				}
				fileSink=gst_element_factory_make("filesink","disk_sink");
				if(fileSink==NULL) {
					errors->setError(options->getFileList()[currentFile],ERROR_GST_INTERNAL);
					abortedOperation(QFileInfo((options->getFileList())[currentFile]).size());
					return;
				}
				gst_element_set(fileSink, "location", outputFile.ascii(), NULL);
				break;
			}
		}
		applyOptions();

		gst_bin_add_many(GST_BIN(pipeline),fileSrc,spider,encoder,fileSink,NULL);
		gst_element_link(fileSrc,spider);
		GstCaps *filter=gst_caps_new_simple("audio/x-raw-int",NULL);
		gst_element_link_filtered(spider, encoder, filter);
		gst_element_link(encoder, fileSink);
 		
		int size=QFile((options->getFileList())[currentFile]).size();
		currentProgress->setTotalSteps(size);
		currentProgress->setProgress(0);
		
		gst_element_set_state(pipeline, GST_STATE_PLAYING);
		
		timer->start(500,false);
		chrono.restart();

	}
	else {
		converting=false;
		finished=true;
		emit titleChanged("");
		emit workDone();
	}
}

void ConvertAdvance::eosCb( GstElement *, ConvertAdvance *win ) {
	//fill the progressbar
	win->currentProgress->setProgress(win->currentProgress->totalSteps());
	win->totalProgress->setProgress(win->currentTotalSize + win->currentProgress->totalSteps());
	win->currentTotalSize+=win->currentProgress->totalSteps();
	
	//copy file tags if so requested and possible
	if(win->options->getMisc(Options::OPTION_TRANSFER_TAGS) && win->options->getFormat()!=wav && win->options->getInputFormat()!=wav) {
		TagLib::FileRef in((win->options->getFileList())[win->currentFile].ascii());
		TagLib::FileRef out(win->outputFile.ascii());
// 		TagLib::Tag::duplicate(in.tag(),out.tag(),true);
		TagLib::Tag *inTag=in.tag();
		TagLib::Tag *outTag=out.tag();
		
		if(!inTag->isEmpty()) {
			outTag->setTitle(inTag->title());
			outTag->setArtist(inTag->artist());
			outTag->setAlbum(inTag->album());
			outTag->setComment(inTag->comment());
			outTag->setGenre(inTag->genre());
			outTag->setYear(inTag->year());
			outTag->setTrack(inTag->track());
		}
	}

	//skip to next file
	++(win->currentFile);
	
	//clear the pipeline
	gst_object_unref(GST_OBJECT(win->pipeline));
	
	//stop the timer
	win->timer->stop();
	
	//start a new file conversion
	win->processFile();
}

void ConvertAdvance::tick( ) {
	gint64 bytes;
	GstFormat format=GST_FORMAT_BYTES;
	if(gst_element_query(fileSrc,GST_QUERY_POSITION,&format,&bytes)) {
		//get processed bytes size
		int pos=static_cast<int>(bytes);
		currentProgress->setProgress(pos);
		totalProgress->setProgress(currentTotalSize+pos);
		
		//processing speed
		QString timeRep;
		int time=chrono.elapsed();
		char unit=' ';
		
		//current file
		int speed=pos/time;
		int remainingMillisec=currentProgress->totalSteps()/speed-time;
		timeRep=i18n("Remaining: ");
		if(remainingMillisec/3600000!=0) { //hours
			if(remainingMillisec/3600000<10)
				timeRep+="0";
			timeRep+=QString::number(remainingMillisec/3600000);
			timeRep+=":";
			remainingMillisec=remainingMillisec%3600000;
			unit='h';
		}
		if(remainingMillisec/60000!=0) { //minutes
			if(remainingMillisec/60000<10)
				timeRep+="0";
			timeRep+=QString::number(remainingMillisec/60000);
			timeRep+=":";
			remainingMillisec=remainingMillisec%60000;
			if(unit==' ')
				unit='m';
		}
		if(remainingMillisec/1000!=0) { //seconds
			if(remainingMillisec/1000<10)
				timeRep+="0";
			timeRep+=QString::number(remainingMillisec/1000);
			if(unit==' ')
				unit='s';
		}
		else
			timeRep+="0 s";
		timeRep+=" ";
		timeRep+=unit;
		currentTime->setText(timeRep);
		
		//total file
		unit=' ';
		remainingMillisec=totalProgress->totalSteps()/speed-time;
		timeRep=i18n("Remaining: ");
		if(remainingMillisec/3600000!=0) { //hours
			if(remainingMillisec/3600000<10)
				timeRep+="0";
			timeRep+=QString::number(remainingMillisec/3600000);
			timeRep+=":";
			remainingMillisec=remainingMillisec%3600000;
			unit='h';
		}
		if(remainingMillisec/60000!=0) { //minutes
			if(remainingMillisec/60000<10)
				timeRep+="0";
			timeRep+=QString::number(remainingMillisec/60000);
			timeRep+=":";
			remainingMillisec=remainingMillisec%60000;
			if(unit==' ')
				unit='m';
		}
		if(remainingMillisec/1000!=0) { //seconds
			if(remainingMillisec/1000<10)
				timeRep+="0";
			timeRep+=QString::number(remainingMillisec/1000);
			if(unit==' ')
				unit='s';
		}
		else
			timeRep+="0 s";
		timeRep+=" ";
		timeRep+=unit;
		totalTime->setText(timeRep);
		
		emit titleChanged(totalProgress->progressString()+" ["+ timeRep +"]");
	}
}

QString ConvertAdvance::createOutputFilename( ) {
	QString toReturn=QFileInfo((options->getFileList())[currentFile]).fileName();
	switch(options->getFormat()) {
		case mp3:
			toReturn.replace(toReturn.length()-4,4,".mp3");
			break;
		case ogg:
			toReturn.replace(toReturn.length()-4,4,".ogg");
			break;
		case wav:
			toReturn.replace(toReturn.length()-4,4,".wav");
			break;
	}
	
	toReturn.prepend(options->getOutputDir()+"/");
	
	if((QFileInfo(toReturn).exists())) {
		errors->setError(toReturn,ERROR_FILE_EXISTS);
		return QString("");
	}
	
	return toReturn;
}

void ConvertAdvance::abortedOperation(int skip) {
	currentProgress->setProgress(0);
	totalProgress->setProgress(currentTotalSize + skip);
	currentTotalSize+=skip;
	
	++currentFile;
	
	processFile();
}

Error *ConvertAdvance::getErrors( ) {
	return errors;
}

void ConvertAdvance::applyOptions( ) {
	switch(options->getFormat()) {
		case mp3: {
			if(options->getBitrate()==choose)  //constant bitrate, user-chosen
				g_object_set(G_OBJECT(encoder),"bitrate",options->getBitrateValue(),NULL);
			// Not working. I have to figure it out how to set gstreamer options. Any idea? write me! Antonio.Fasolato@poste.it
/*			if(options->getBitrate()==average)  //constant bitrate, user-chosen
				g_object_set(G_OBJECT(encoder),"bitrate",options->getBitrateValue(),NULL);
			if(options->getBitrate()==variable) { //constant bitrate, user-chosen
				g_object_set(G_OBJECT(encoder),"vbr",vbr_rh,NULL);
				g_object_set(G_OBJECT(encoder),"vbr_quality",options->getBitrateValue(),NULL);
			}*/
			break;
		}
		case ogg: {
			if(options->getBitrate()==choose) {  //constant bitrate, user-chosen
				float quality=static_cast<float>(options->getBitrateValue())/10;
				g_object_set(G_OBJECT(encoder),"quality",quality,NULL);
			}
			break;
		}
		case wav: {
			break;
		}
	}
}

bool ConvertAdvance::cancel( ) {
	if(converting) {
		int answer=KMessageBox::questionYesNo(this,i18n("<h1>Abort?</h1><p>Are you shure you want to stop the conversion?</p><p>This will abort all operations and close this application.</p>"), i18n("Abort conversion?"));
		return (answer==KMessageBox::Yes);
	}
	else
		return true;
}

/*$SPECIALIZATION$*/


#include "convertadvance.moc"

